import logging
import json
import os

from aperturedb.Connector import Connector
from aperturedb.ConnectorRest import ConnectorRest
from aperturedb import ProgressBar
from aperturedb.ParallelQuery import execute_batch
from aperturedb.Configuration import Configuration
from aperturedb.cli.configure import ls
from aperturedb.cli.console import console


logger = logging.getLogger(__name__)

DESCRIPTOR_CLASS = "_Descriptor"
DESCRIPTOR_CONNECTION_CLASS = "_DescriptorSetToDescriptor"

DEFAULT_METADATA_BATCH_SIZE = 100_000


def __create_connector(configuration: Configuration):
    if configuration.use_rest:
        connector = ConnectorRest(
            host=configuration.host,
            port=configuration.port,
            user=configuration.username,
            password=configuration.password,
            config=configuration)
    else:
        connector = Connector(
            host=configuration.host,
            port=configuration.port,
            user=configuration.username,
            password=configuration.password,
            config=configuration)
    console.log(f"Connected Using:")
    console.log(configuration)
    return connector


def create_connector():
    """
    **Create a connector to the database.**

    Args:
        None

    Returns:
        Connector: The connector to the database.
    """
    all_configs = ls()

    env_config = os.environ.get("APERTUREDB_CONFIG")
    ac = all_configs["active"]
    config = all_configs["local"][ac] if "local" in all_configs else all_configs["global"][ac]

    if env_config is not None:
        # TODO test me.
        config = all_configs["global"][env_config] if env_config in all_configs["global"] else all_configs["local"][env_config]
        return __create_connector(config)
    # Then check if the local config has active
    else:
        return __create_connector(config)


class Utils(object):
    """
    **A bunch of helper methods to get information from aperturedb.**

    Args:
        object (Connector): The underlying Connector.
    """

    def __init__(self, connector: Connector, verbose=False):

        self.connector = connector.create_new_connection()
        self.verbose = verbose

    def __repr__(self):

        return self.status()

    def print(self, str):

        if self.verbose:
            print(str)

    def execute(self, query, blobs=[], success_statuses=[0]):
        try:
            rc, r, b = execute_batch(
                query, blobs, self.connector, success_statuses=success_statuses)
        except BaseException as e:
            logger.error(self.connector.get_last_response_str())
            raise e

        if rc != 0:
            raise Exception(f"query failed with result {rc}")

        return r, b

    def status(self):

        q = [{"GetStatus": {}}]

        self.execute(q)

        return self.connector.get_last_response_str()

    def print_schema(self, refresh=False):

        if refresh:
            logger.warning("get_schema: refresh no longer needed.")
            logger.warning("get_schema: refresh will be deprecated.")
            logger.warning("Please remove 'refresh' parameter.")

        self.get_schema()
        self.connector.print_last_response()

    def get_schema(self, refresh=False):

        if refresh:
            logger.warning("get_schema: refresh no longer needed.")
            logger.warning("get_schema: refresh will be deprecated.")
            logger.warning("Please remove 'refresh' parameter.")

        query = [{
            "GetSchema": {
            }
        }]

        res, _ = self.connector.query(query)

        schema = {}

        try:
            schema = res[0]["GetSchema"]
        except BaseException as e:
            logger.error(self.connector.get_last_response_str())
            raise e

        return schema

    def _object_summary(self, name, object):

        total_elements = object["matched"]

        print(f"{name.ljust(20)}")
        if "src" in object:
            print(f"  {object['src']} ====> {object['dst']}")

        print(f"  Total elements: {total_elements}")

        p = object["properties"]

        if p is None:
            return total_elements

        max = 0
        for k in p:
            max = len(k) if max < len(k) else max
        max += 1

        for k in p:
            i = w = " "
            i = "I" if p[k][1] else i

            # Warning if there are some properties not present in all elements
            w = "!" if p[k][0] != total_elements else w
            # Warning if there is a property with "id" in the name not indexes
            w = "!" if "id" in k and not p[k][1] else w
            print(f"{i} {w} {p[k][2].ljust(8)} |"
                  f" {k.ljust(max)} | {str(p[k][0]).rjust(9)} "
                  f"({int(p[k][0]/total_elements*100.0)}%)")

        return total_elements

    def summary(self):

        r = self.get_schema()
        s = json.loads(self.status())[0]["GetStatus"]
        version = s["version"]
        status  = s["status"]
        info    = s["info"]

        if r["entities"] == None:
            total_entities = 0
            entities_classes = []
        else:
            total_entities   = r["entities"]["returned"]
            entities_classes = [c for c in r["entities"]["classes"]]

        if r["connections"] == None:
            total_connections = 0
            connections_classes = []
        else:
            total_connections   = r["connections"]["returned"]
            connections_classes = [c for c in r["connections"]["classes"]]

        print(f"================== Summary ==================")
        print(f"Database: {self.connector.host}")
        print(f"Version: {version}")
        print(f"Status:  {status}")
        print(f"Info:    {info}")
        print(f"------------------ Entities -----------------")
        print(f"Total entities types:    {total_entities}")
        total_nodes = 0
        for c in entities_classes:
            total_nodes += self._object_summary(c, r["entities"]["classes"][c])

        print(f"---------------- Connections ----------------")
        print(f"Total connections types: {total_connections}")
        total_edges = 0
        for c in connections_classes:
            total_edges += self._object_summary(c,
                                                r["connections"]["classes"][c])

        print(f"------------------ Totals -------------------")
        print(f"Total nodes: {total_nodes}")
        print(f"Total edges: {total_edges}")
        print(f"=============================================")

    def _create_index(self, index_type, class_name, property_key):

        q = [{
            "CreateIndex": {
                "index_type":    index_type,
                "class":         class_name,
                "property_key":  property_key,
            }
        }]

        try:
            self.execute(q, success_statuses=[0, 2])
        except:
            return False

        return True

    def _remove_index(self, index_type, class_name, property_key):

        q = [{
            "RemoveIndex": {
                "index_type":   index_type,
                "class":        class_name,
                "property_key": property_key,
            }
        }]

        try:
            self.execute(q)
        except:
            return False

        return True

    def create_entity_index(self, class_name, property_key, property_type=None):

        if property_type is not None:
            logger.warning(f"create_entity_index ignores 'property_type'")
            logger.warning(f"'property_type' will be deprecated in the future")
            logger.warning(f"Please remove 'property_type' parameter")

        return self._create_index("entity", class_name, property_key)

    def create_connection_index(self, class_name, property_key, property_type=None):

        if property_type is not None:
            logger.warning(f"create_connection_index ignores 'property_type'")
            logger.warning(f"'property_type' will be deprecated in the future")
            logger.warning(f"Please remove 'property_type' parameter")

        return self._create_index("connection", class_name, property_key)

    def remove_entity_index(self, class_name, property_key):

        return self._remove_index("entity", class_name, property_key)

    def remove_connection_index(self, class_name, property_key):

        return self._remove_index("connection", class_name, property_key)

    def count_images(self, constraints={}):

        q = [{
            "FindImage": {
                "blobs": False,
                "results": {
                    "count": True,
                }
            }
        }]

        if constraints:
            q[0]["FindImage"]["constraints"] = constraints

        res, _ = self.execute(q)
        total_images = res[0]["FindImage"]["count"]

        return total_images

    def get_uniqueids(self, object_type, constraints={}):

        q = [{
            "FindEntity": {
                "with_class": object_type,
                "batch": {},
                "results": {
                    "list": ["_uniqueid"],
                }
            }
        }]

        if constraints:
            q[0]["FindEntity"]["constraints"] = constraints

        ids = []

        res, _ = self.execute(q)
        total_elements = res[0]["FindEntity"]["batch"]["total_elements"]

        batch_size = DEFAULT_METADATA_BATCH_SIZE
        iterations = total_elements // batch_size
        reminder   = total_elements % batch_size

        if iterations == 0 and reminder > 0:
            iterations = 1

        pb = ProgressBar.ProgressBar()

        for i in range(iterations):

            batch = {
                "batch_size": batch_size,
                "batch_id": i
            }

            q[0]["FindEntity"]["batch"] = batch

            res, _ = self.execute(q)
            ids += [element["_uniqueid"]
                    for element in res[0]["FindEntity"]["entities"]]

            if self.verbose:
                pb.update(i / iterations)

        if self.verbose:
            pb.update(1)  # For the end of line

        return ids

    def get_images_uniqueids(self, constraints={}):

        return self.get_uniqueids("_Image", constraints)

    def count_bboxes(self, constraints=None):
        # The default params in python functions should not be
        # mutable objects.
        # It can lead to https://docs.python-guide.org/writing/gotchas/#mutable-default-arguments
        q = [{
            "FindBoundingBox": {
                "blobs": False,
                "results": {
                    "count": True,
                }
            }
        }]

        if constraints:
            q[0]["FindBoundingBox"]["constraints"] = constraints

        res, _ = self.execute(q)
        total_connections = res[0]["FindBoundingBox"]["count"]

        return total_connections

    def count_entities(self, entity_class, constraints=None):

        q = [{
            "FindEntity": {
                "with_class": entity_class,
                "results": {
                    "count": True,
                }
            }
        }]

        if constraints:
            q[0]["FindEntity"]["constraints"] = constraints

        res, _ = self.execute(q)
        total_entities = res[0]["FindEntity"]["count"]

        return total_entities

    def count_connections(self, connections_class, constraints=None):

        q = [{
            "FindConnection": {
                "with_class": connections_class,
                "results": {
                    "count": True,
                }
            }
        }]

        if constraints:
            q[0]["FindConnection"]["constraints"] = constraints

        res, _ = self.execute(q)
        total_connections = res[0]["FindConnection"]["count"]

        return total_connections

    def add_descriptorset(self, name, dim, metric="L2", engine="FaissFlat"):

        query = [{
            "AddDescriptorSet": {
                "name":       name,
                "dimensions": dim,
                "metric":     metric,
                "engine":     engine
            }
        }]

        try:
            self.execute(query)
        except:
            return False

        return True

    def count_descriptorsets(self):

        q = [{
            "FindDescriptorSet": {
                "results": {
                    "count": True,
                }
            }
        }]

        res, _ = self.query(q)
        total_descriptor_sets = res[0]["FindDescriptorSet"]["count"]

        return total_descriptor_sets

    def get_descriptorset_list(self):

        q = [{
            "FindDescriptorSet": {
                "results": {
                    "list": ["_name"],
                }

            }
        }]

        sets = []
        res, _ = self.execute(q)
        if res[0]["FindDescriptorSet"]["returned"] > 0:
            sets = [ent["_name"]
                    for ent in res[0]["FindDescriptorSet"]["entities"]]

        return sets

    def remove_descriptorset(self, set_name):

        q = [{
            "FindDescriptorSet": {
                "_ref": 1,
                "with_name": set_name,

            }
        }, {
            "DeleteDescriptorSet": {
                "ref": 1
            }
        }]

        try:
            self.execute(q)
        except:
            return False

        return True

    def _remove_objects(self, type, class_name, batch_size):

        if type == "entities":
            cmd = "Entity"
            count = self.count_entities(class_name)
        elif type == "connections":
            cmd = "Connection"
            count = self.count_connections(class_name)
        else:
            raise ValueError("Type must be either 'entities' or 'connections'")

        total = count

        pb = ProgressBar.ProgressBar()

        find = "Find" + cmd
        dele = "Delete" + cmd
        while count > 0:

            q = [{
                find: {
                    "_ref": 1,
                    "with_class": class_name,
                    "results": {
                        "limit": batch_size
                    }
                }
            }, {
                dele: {
                    "ref": 1
                }
            }]

            try:
                self.execute(q)
            except:
                return False

            self.print(
                f"Delete transaction of {batch_size} elements took: {self.connector.get_last_query_time()}")

            count -= batch_size

            if self.verbose:
                pb.update(abs((count - total) / total))

        return True

    def remove_entities(self, class_name, batched=False, batch_size=10000):

        # We used to batch because removing a large number of object did not work ok.
        # Now it seems to work ok, so we don't batch anymore.
        # We keep the option in case we need to batch, but not default.
        if batched:
            return self._remove_objects("entities", class_name, batch_size)

        query = [{
            "DeleteEntity": {
                "with_class": class_name
            }
        }]

        try:
            r, _ = self.execute(query)
            count = r[0]["DeleteEntity"]["count"]
            self.print(f"Deleted {count} entities")
            self.print(
                f"Delete transaction took: {self.connector.get_last_query_time()}")
        except:
            logger.error("Could not get count of deleted entities")
            return False

        return True

    def remove_connections(self, class_name, batched=False, batch_size=10000):

        # We used to batch because removing a large number of object did not work ok.
        # Now it seems to work ok, so we don't batch anymore.
        # We keep the option in case we need to batch, but not default.
        if batched:
            return self._remove_objects("connections", class_name, batch_size)

        query = [{
            "DeleteConnection": {
                "with_class": class_name
            }
        }]

        try:
            r, _ = self.execute(query)
            count = r[0]["DeleteConnection"]["count"]
            self.print(f"Deleted {count} connections")
            self.print(
                f"Delete transaction took: {self.connector.get_last_query_time()}")
        except:
            logger.error("Could not get count of deleted connections")
            return False

        return True

    def remove_all_descriptorsets(self):

        self.print("Removing indexes...")

        idx_props = self.get_indexed_props(DESCRIPTOR_CLASS)

        for idx in idx_props:
            self.remove_entity_index(DESCRIPTOR_CLASS, idx)

        self.print("Done removing indexes.")

        self.print("Removing connections...")
        if not self.remove_connections(DESCRIPTOR_CONNECTION_CLASS):
            print("Error removing connections.")
            return False
        self.print("Done removing connections.")

        self.print("Removing descriptors...")
        if not self.remove_entities(DESCRIPTOR_CLASS):
            print("Error removing descriptors.")
            return False
        self.print("Done removing descriptors.")

        sets = self.get_descriptorset_list()

        self.print("Removing sets...")
        for s in sets:
            self.print("Removing {}...".format(s))
            self.remove_descriptorset(s)

        self.print("Done removing sets.")

    def get_indexed_props(self, class_name, type="entities"):
        """
        Returns all the indexed properties for a given class.
        """

        if type not in ["entities", "connections"]:
            raise ValueError("Type must be either 'entities' or 'connections'")

        schema = self.get_schema()

        try:
            indexed_props = schema[type]["classes"][class_name]["properties"]
            indexed_props = [
                k for k in indexed_props.keys() if indexed_props[k][1]]
        except:
            indexed_props = []

        return indexed_props

    def count_descriptors_in_set(self, set_name):

        total = -1

        q = [{
            "FindDescriptorSet": {
                "_ref": 1,
                "with_name": set_name,
            }
        }, {
            "FindDescriptor": {
                "set": set_name,
                "is_connected_to": {
                    "ref": 1,
                    "connection_class": "_DescriptorSetToDescriptor"
                },
                "results": {
                    "count": True
                }
            }
        }]

        res, _ = self.execute(q)
        total = res[1]["FindDescriptor"]["count"]

        return total

    def remove_all_objects(self):

        cmd = {"constraints": {"_uniqueid": ["!=", "0.0.0"]}}

        transaction = [
            {"DeleteImage": cmd},
            {"DeleteVideo": cmd},
            {"DeleteBlob": cmd},
            # DeleteDescriptorSet also deletes the descriptors
            {"DeleteDescriptorSet": cmd},

            # All the following should be deleted automatically once the
            # related objects are deleted.
            # We keep them here until ApertureDB fully implements this.
            {"DeleteBoundingBox": cmd},
            {"DeletePolygon": cmd},
            {"DeleteFrame": cmd},
            {"DeleteEntity": cmd},
            {"GetSchema": {"refresh": True}}
        ]

        try:
            response, _ = self.execute(transaction)
            schema = response[-1]["GetSchema"]
            check_keys = ["connections", "entities"]
            return schema["status"] == 0 and \
                all(map(lambda k: schema[k] is None, check_keys))
        except BaseException as e:
            logger.exception(e)

        return False



import argparse
import random
import math
import os

from datetime import datetime

import pandas as pd
import numpy as np
from generateImages import ImageGenerator

from itertools import product


def remove_prefix(text, prefix):
    if text.startswith(prefix):
        return text[len(prefix):]
    else:
        return text


def generate_person_csv(multiplier):

    names     = ["James", "Luis", "Sole", "Maria",
                 "Tom", "Xavi", "Dimitris", "King"]
    lastnames = ["Ramirez", "Berlusconi", "Copola", "Tomson", "Ferro"]
    names     = names     * multiplier
    lastnames = lastnames * multiplier

    persons  = list(product(names, lastnames))

    entity   = ["Person" for x in range(len(persons))]
    ids      = random.sample(range(1000000000), len(persons))
    age      = [int(100 * random.random()) for i in range(len(persons))]
    height   = [float(200 * random.random()) for i in range(len(persons))]
    dog      = [x > 100 for x in height]
    birth    = [datetime.now().isoformat() for x in range(len(persons))]

    df = pd.DataFrame(persons, columns=['name', 'lastname'])
    df["EntityClass"] = entity
    df = df.reindex(["EntityClass", "name", "lastname"], axis=1)
    df["id"]       = ids
    df["age"]      = age
    df["height"]   = height
    df["has_dog"]  = dog
    df["date:dob"] = birth
    df["constraint_id"] = ids

    df = df.sort_values("id")

    df.to_csv("input/persons.adb.csv", index=False)

    return df


def generate_blobs_csv():

    path       = "blobs/"
    blob_paths = [path + str(x).zfill(4) + ".blob" for x in range(20)]
    license    = [x for x in range(10)]
    blobs      = list(product(blob_paths, license))

    # generate the blobs
    arr = [1.5, 2.4, 3.3, 5.5, 9.9, 111.12, 1000.20]
    nparr = np.array(arr)
    for blob in blob_paths:
        with open(os.path.join("input", blob), 'wb') as blobfile:
            nparr.tofile(blobfile)

    entity   = ["segmentation" for x in range(len(blobs))]
    ids      = random.sample(range(1000000000), len(blobs))

    df = pd.DataFrame(blobs, columns=['filename', 'license'])
    df                  = df.reindex(["filename", "license"], axis=1)
    df["id"]            = ids
    df["constraint_id"] = ids

    df = df.sort_values("id")

    df.to_csv("input/blobs.adb.csv", index=False)

    return df


def generate_images_csv(multiplier):

    multiplier = multiplier // 2
    path    = "images/"
    imgs    = [path + str(x).zfill(4) +
               ".jpg" for x in range(200)] * multiplier
    license = [x for x in range(10)] * multiplier

    images  = list(product(imgs, license))

    ids      = random.sample(range(1000000000), len(images))
    age      = [int(100 * random.random()) for i in range(len(images))]
    height   = [float(200 * random.random()) for i in range(len(images))]
    dog      = [x > 100 for x in height]
    date_cap = [datetime.now().isoformat() for x in range(len(images))]

    df = pd.DataFrame(images, columns=['filename', 'license'])
    df["id"]       = ids
    df["age"]      = age
    df["height"]   = height
    df["has_dog"]  = dog
    df["date:date_captured"] = date_cap
    df["constraint_id"] = ids

    df = df.sort_values("id")

    df.to_csv("input/images.adb.csv", index=False)

    return df


def generate_http_images_csv(ip_file_csv):

    images    = pd.read_csv(ip_file_csv, sep=",", header=None)

    ids      = random.sample(range(1000000000), len(images))
    age      = [int(100 * random.random()) for i in range(len(images))]
    height   = [float(200 * random.random()) for i in range(len(images))]
    license  = [x for x in range(len(images))]

    df = pd.DataFrame()
    df['url']      = images
    df["urlid"]    = ids
    df['license']  = license
    df["age"]      = age
    df["height"]   = height
    df["constraint_urlid"] = ids

    df = df.sort_values("urlid")

    df.to_csv("input/http_images.adb.csv", index=False)

    return df


def generate_s3_images_csv(ip_file_csv):

    images    = pd.read_csv(ip_file_csv, sep=",", header=None)

    ids      = random.sample(range(1000000000), len(images))
    age      = [int(100 * random.random()) for i in range(len(images))]
    height   = [float(200 * random.random()) for i in range(len(images))]
    license  = [x for x in range(len(images))]

    df = pd.DataFrame()
    df['s3_url']   = images
    df["id"]       = ids
    df['license']  = license
    df["age"]      = age
    df["height"]   = height
    df["constraint_id"] = ids

    df = df.sort_values("id")

    df.to_csv("input/s3_images.adb.csv", index=False)

    return df


def generate_gs_images_csv(ip_file_csv):

    images    = pd.read_csv(ip_file_csv, sep=",", header=None)

    ids      = random.sample(range(1000000000), len(images))
    age      = [int(100 * random.random()) for i in range(len(images))]
    height   = [float(200 * random.random()) for i in range(len(images))]
    license  = [x for x in range(len(images))]

    df = pd.DataFrame()
    df['gs_url']   = images
    df["id"]       = ids
    df['license']  = license
    df["age"]      = age
    df["height"]   = height
    df["constraint_id"] = ids

    df = df.sort_values("id")

    df.to_csv("input/gs_images.adb.csv", index=False)

    return df


def generate_connections_csv(persons, images):

    connections  = list(product(images["id"][::100], persons["id"][::100]))

    connect   = ["has_image" for x in range(len(connections))]
    confidence  = [random.random() for x in range(len(connections))]

    df = pd.DataFrame(connections, columns=['_Image@id', 'Person@id'])
    df["ConnectionClass"] = connect
    df = df.reindex(["ConnectionClass", '_Image@id', 'Person@id'], axis=1)

    df["confidence"]      = confidence

    df.to_csv("input/connections-persons-images.adb.csv", index=False)


def generate_bboxes_csv(images):

    images  = images["id"]

    x   = [0 for x in range(len(images))]
    y   = [0 for x in range(len(images))]
    w   = [100 for x in range(len(images))]
    h   = [150 for x in range(len(images))]

    labels = ["dog", "cat", "catdog", "rocko", "philip", "froghead", "drog"]

    confidence  = [random.random() for x in range(len(images))]
    label       = [labels[math.floor(random.random() * len(labels))]
                   for x in range(len(images))]

    df = pd.DataFrame(images, columns=['id'])

    df["x_pos"] = x
    df["y_pos"] = y
    df["width"] = w
    df["height"] = h
    df["confidence"] = confidence
    df["label"]      = label

    df = df.sort_values("id")
    df.to_csv("input/bboxes.adb.csv", index=False)


def generate_bboxes_constraints_csv(images):
    image_id = images.iloc[0]["id"]
    num_boxes = 3

    x   = [0 for x in range(num_boxes)]
    y   = [0 for x in range(num_boxes)]
    w   = [100 for x in range(num_boxes)]
    h   = [150 for x in range(num_boxes)]

    labels = ["dog", "cat", "catdog", "rocko", "philip", "froghead", "drog"]

    confidence  = [random.random() for x in range(num_boxes)]
    label       = [labels[math.floor(random.random() * num_boxes)]
                   for x in range(num_boxes)]

    df = pd.DataFrame([{
        "id": image_id,
    } for _ in range(num_boxes)])

    df["x_pos"] = x
    df["y_pos"] = y
    df["width"] = w
    df["height"] = h
    df["confidence"] = confidence
    df["label"]      = label
    df["box_id"] = [123] * num_boxes
    df["constraint_box_id"] = [123] * num_boxes

    df = df.sort_values("id")
    df.to_csv("input/bboxes-constraints.adb.csv", index=False)


def generate_descriptors(images, setname, dims):
    filename = "input/" + setname + "_desc.npy"
    images  = images["id"]
    descriptors = np.random.rand(len(images), dims)
    np.save(filename, descriptors)

    filename = setname + "_desc.npy"

    labels = ["dog", "cat", "catdog", "rocko", "philip", "froghead", "drog"]
    confidence  = [random.random() for x in range(len(images))]
    label       = [labels[math.floor(random.random() * len(labels))]
                   for x in range(len(images))]

    df = pd.DataFrame(images, columns=['id'])

    df["filename"] = [filename for x in range(len(images))]
    df["index"]    = [i for i in range(len(images))]
    df["set"]      = [setname for x in range(len(images))]
    df["id"]       = [i for i in range(len(images))]
    df["label"]      = label
    df["confidence"] = confidence

    df = df.sort_values("id")
    df = df.reindex(columns=["filename", "index",
                    "set", "id", "label", "confidence"])
    df.to_csv("input/" + setname + ".adb.csv", index=False)


def generate_descriptorset(names, dims):

    metrics = ["L2", "IP", "CS", ["CS", "IP"], ["L2", "CS", "IP"]]
    engines = ["FaissIVFFlat", "FaissFlat", ["FaissIVFFlat", "FaissFlat"]]

    df = pd.DataFrame()

    df["name"]       = names
    df["dimensions"] = dims
    df["engine"]     = [random.sample(engines, 1)[0] for i in names]
    df["metric"]     = [random.sample(metrics, 1)[0] for i in names]

    df.to_csv("input/descriptorset.adb.csv", index=False)


def generate_update_person():
    # generate 3 testing csvs for update testing
    # cvs 1 - base load with version.
    df = pd.read_csv("input/persons.adb.csv")

    # set version column to 1.
    version_id = [1 for x in range(len(df))]
    df["updateif_<version_id"] = version_id
    df["version_id"] = version_id
    df.to_csv("input/persons-update.adb.csv", index=False)

    # csv 2 - modified data, but same version ( will cause no updates )
    df['age'] = df['age'].apply(lambda age: age + 200)
    df['version_id'] = df['version_id'].apply(lambda ver: ver + 1)
    df.to_csv("input/persons-update-oldversion.adb.csv", index=False)

    # csv 3 - modified data, and version ( will update )
    # update in database if version < 2.
    df['updateif_<version_id'] = df['version_id']
    df.to_csv("input/persons-update-newversion.adb.csv", index=False)

    # generate updateif to test > comparator and partial update
    df = pd.read_csv("input/persons.adb.csv")
    df['age'] = df['age'].apply(lambda age: age + 200)
    # change if age in database is > 30
    df['updateif_>age'] = [30 for x in range(len(df))]
    df = df.to_csv("input/persons-update-olderage.adb.csv", index=False)


def generate_partial_load():

    df = pd.read_csv("input/persons.adb.csv")
    base = df.head(10)
    base.to_csv("input/persons-exist-base.adb.csv", index=False)
    # causes 3 overlaps
    overlapped = df.head(12).tail(5)
    overlapped.to_csv("input/persons-some-exist.adb.csv", index=False)


def generate_update_images(multiplier):

    image_count = 5
    licence_count = 2
    multiplier = multiplier // 2
    path    = "images/"
    imgs    = [path + "number_" + str(i).zfill(4) +
               ".png" for i in range(image_count)] * multiplier
    license = [x for x in range(licence_count)] * multiplier

    images  = list(product(imgs, license))
    duplicate_count = len(images) // 10

    ids      = random.sample(range(1000000000), len(images))
    age      = [int(100 * random.random()) for i in range(len(images))]
    height   = [float(200 * random.random()) for i in range(len(images))]
    dog      = [x > 100 for x in height]
    date_cap = [datetime.now().isoformat() for x in range(len(images))]
    version_id = [1 for x in range(len(images))]

    df = pd.DataFrame(images, columns=['filename', 'license'])
    df["id"]       = ids
    df["age"]      = age
    df["height"]   = height
    df["has_dog"]  = dog
    df["date:date_captured"] = date_cap
    df["constraint_id"] = ids
    df["version_id"] = version_id
    df["updateif_<version_id"] = version_id

    df = df.sort_values("id")

    updates = df.head(duplicate_count).copy()
    updated_version_id = [2 for i in range(duplicate_count)]
    updates["version_id"] = updated_version_id
    updates["updateif_<version_id"] = updated_version_id
    df = pd.concat([df, updates])

    df.to_csv("./input/images_update_and_add.adb.csv", index=False)


def generate_newest_images(multiplier):
    image_count = 5
    licence_count = 2
    multiplier = multiplier // 2
    path    = "images/"
    img_ids = [i for i in range(image_count)] * multiplier
    def filegen(file_num): return path + "number_" + \
        str(file_num).zfill(4) + ".png"
    license = [x for x in range(licence_count)] * multiplier

    images  = list(product(img_ids, license))
    prop_change_count = len(images) // 10

    ids      = random.sample(range(1000000000), len(images))
    age      = [int(100 * random.random()) for i in range(len(images))]
    height   = [float(200 * random.random()) for i in range(len(images))]
    dog      = [x > 100 for x in height]
    date_cap = [datetime.now().isoformat() for x in range(len(images))]
    version_id = [1 for x in range(len(images))]
    empty_column = ["" for x in range(len(images))]

    df = pd.DataFrame(images, columns=['img_id', 'license'])
    # we want filename to be first column
    df.insert(0, "filename", df['img_id'].apply(filegen))
    df["id"]       = ids
    df["age"]      = age
    df["height"]   = height
    df["has_dog"]  = dog
    df["date:date_captured"] = date_cap
    df["constraint_id"] = ids
    df["version_id"] = version_id
    df["gen_blobsha1_imagesha"] = empty_column
    df["updateif_<version_id"] = version_id

    df = df.sort_values("id")

    # temporary
    df = df.head(4).copy()
    prop_change_count = 1
    blob_change_count = 1

    prop_updates = df.head(prop_change_count).copy()
    prop_updates["version_id"] = prop_updates["version_id"].apply(
        lambda id: id + 1)
    prop_updates["updateif_<version_id"] = prop_updates["version_id"]
    prop_updates["age"] = prop_updates["age"].apply(lambda age: age + 1)

    blob_updates = df.head(prop_change_count + 1).tail(1).copy()
    blob_updates["version_id"] = blob_updates["version_id"].apply(
        lambda id: id + 2)
    blob_updates["updateif_<version_id"] = blob_updates["version_id"]
    blob_updates["age"] = blob_updates["age"].apply(lambda age: age + 2)
    blob_updates["img_id"] = blob_updates["img_id"].apply(lambda id: id + 1)
    blob_updates["filename"] = blob_updates["img_id"].apply(filegen)
    df = pd.concat([df, prop_updates, blob_updates])
    print(df)
    df = df.drop(columns=["img_id"])

    df.to_csv("./input/images_newest_blobs.adb.csv", index=False)


def generate_update_image(multiplier):
    # generate base load
    # generate images
    image_count = 100
    img_gen = ImageGenerator(count=image_count, output="input/images/update_images_%%.png", image_type="png",
                             size=(256, 256), manifest="input/update_image_list.csv")
    img_gen.run()
    img_df = pd.read_csv("input/update_image_list.csv", header=None)
    licence_count = 2
    multiplier = multiplier // 2
    img_ids = [i for i in range(image_count)] * multiplier
    license = [x for x in range(licence_count)] * multiplier

    def filemap(file_num):
        return remove_prefix(img_df.iat[file_num % image_count, 0], "input/")
    images  = list(product(img_ids, license))
    prop_change_count = len(images) // 10
    id_range = 500000000
    # id_range=1000000000
    df = pd.DataFrame(images, columns=['img_id', 'license'])
    ids      = random.sample(range(id_range), len(images))
    age      = [int(100 * random.random()) for i in range(len(images))]
    height   = [float(200 * random.random()) for i in range(len(images))]
    dog      = [x > 100 for x in height]
    date_cap = [datetime.now().isoformat() for x in range(len(images))]
    version_id = [1 for x in range(len(images))]

    df = pd.DataFrame(images, columns=['img_id', 'license'])
    # we want filename to be first column
    df.insert(0, "filename", df['img_id'].apply(filemap))
    df["id"]       = ids
    df["age"]      = age
    df["height"]   = height
    df["has_dog"]  = dog
    df["date:date_captured"] = date_cap
    df["constraint_id"] = ids
    df["version_id"] = version_id
    df["updateif_<version_id"] = version_id

    df.to_csv("./input/images_updateif_baseload.adb.csv", index=False)

    # csv 2: original content plus new to verify blob loading with only partial loads, randomize
    id_range_start = id_range
    id_range_end = 1000000000
    new_start = image_count
    new_end = int(image_count * 1.1)
    img_ids = [i for i in range(new_start, new_end)] * multiplier
    images  = list(product(img_ids, license))
    prop_change_count = len(images) // 10
    additional_df = pd.DataFrame(images, columns=['img_id', 'license'])
    ids      = random.sample(range(id_range_start, id_range_end), len(images))
    age      = [int(100 * random.random()) for i in range(len(images))]
    height   = [float(200 * random.random()) for i in range(len(images))]
    dog      = [x > 100 for x in height]
    date_cap = [datetime.now().isoformat() for x in range(len(images))]
    version_id = [1 for x in range(len(images))]

    # we want filename to be first column
    additional_df.insert(0, "filename", df['img_id'].apply(filemap))
    additional_df["id"]       = ids
    additional_df["age"]      = age
    additional_df["height"]   = height
    additional_df["has_dog"]  = dog
    additional_df["date:date_captured"] = date_cap
    additional_df["constraint_id"] = ids
    additional_df["version_id"] = version_id
    additional_df["updateif_<version_id"] = version_id

    additional_df.to_csv(
        "./input/images_updateif_mixednew1.adb.csv", index=False)
    # mix the new records in with the old.
    combined_df = pd.concat([df, additional_df],
                            ignore_index=True).sample(frac=1)

    combined_df.to_csv("./input/images_updateif_mixednew.adb.csv", index=False)


def generate_update_image_fail(multiplier):
    # generate base load, small images.
    image_count = 100
    img_gen = ImageGenerator(count=image_count, output="input/images/update_fail_images_%%.png", image_type="png",
                             size=(32, 32), manifest="input/update_fail_image_list.csv")
    img_gen.run()
    img_df = pd.read_csv("input/update_fail_image_list.csv", header=None)
    licence_count = 2
    multiplier = multiplier // 2
    img_ids = [i for i in range(image_count)] * multiplier
    license = [x for x in range(licence_count)] * multiplier

    def filemap(file_num):
        return remove_prefix(img_df.iat[file_num % image_count, 0], "input/")
    images  = list(product(img_ids, license))
    prop_change_count = len(images) // 10
    id_range = 500000000
    df = pd.DataFrame(images, columns=['img_id', 'license'])
    ids      = random.sample(range(id_range), len(images))
    age      = [int(100 * random.random()) for i in range(len(images))]
    height   = [float(200 * random.random()) for i in range(len(images))]
    dog      = [x > 100 for x in height]
    date_cap = [datetime.now().isoformat() for x in range(len(images))]
    version_id = [1 for x in range(len(images))]

    df.insert(0, "filename", df['img_id'].apply(filemap))
    df["id"]       = ids
    df["age"]      = age
    df["height"]   = height
    df["has_dog"]  = dog
    df["date:date_captured"] = date_cap
    df["constraint_id"] = ids
    df["version_id"] = version_id
    df["updateif_<version_id"] = version_id
    df.to_csv("./input/images_updateif_fail_baseload.adb.csv", index=False)

    # 2nd csv - original data, but new bigger images. update_id will increase, but image will remain the same. ( as blobs are not modifable )
    # 300x300 is chosen to allow twice as large - we also add some more text to ensure the image is more complex ( and thus larger )
    img_gen = ImageGenerator(count=image_count, output="input/images/update_fail_big_images_%%.png", image_type="png",
                             size=(300, 300), manifest="input/update_fail_big_image_list.csv", append_text="_bigimage")
    img_gen.run()
    big_img_df = pd.read_csv(
        "input/update_fail_big_image_list.csv", header=None)
    failing_update = df.copy()

    def big_fail_filemap(file_num):
        return remove_prefix(big_img_df.iat[file_num % image_count, 0], "input/")
    failing_update["filename"] = failing_update["img_id"].apply(
        big_fail_filemap)
    failing_update['version_id'] = failing_update['version_id'].apply(
        lambda ver: ver + 1)
    failing_update["updateif_<version_id"] = failing_update["version_id"]
    failing_update.to_csv(
        "./input/images_updateif_fail_updates.adb.csv", index=False)


def generate_forceimage_load(multiplier):
    # copy from imageupdate.
    # ensure normal load.
    base = pd.read_csv("./input/images_updateif_baseload.adb.csv")
    base.to_csv("./input/images_forceupdate_baseload.adb.csv", index=False)
    # ensure loading with partial new works as expected.
    mixednew = pd.read_csv("./input/images_updateif_mixednew.adb.csv")
    mixednew.to_csv("./input/images_forceupdate_mixednew.adb.csv", index=False)

    # test loading of modified image
    changeimage_base = pd.read_csv(
        "./input/images_updateif_fail_baseload.adb.csv")
    # with a modified id, the id will change, but not the images.
    changeimage_modified = pd.read_csv(
        "./input/images_updateif_fail_updates.adb.csv")

    changeimage_base.to_csv(
        "./input/images_forceupdate_fail_base.adb.csv", index=False)
    changeimage_modified.to_csv(
        "./input/images_forceupdate_fail_updates.adb.csv", index=False)

    # now change load to add column to verify image.
    empty_column = ["" for x in range(len(changeimage_base))]

    # add autogenerated column ( needed to seed the data )
    changeimage_base["prop_blobsha1_imagesha"] = empty_column
    # non should exist in base load, so will not effect anything
    changeimage_base["updateif_blobsha1_imagesha"] = empty_column

    # add autogenerated column to update.
    changeimage_modified["prop_blobsha1_imagesha"] = empty_column
    # required to detect blob change and remove and re-add.
    changeimage_modified["updateif_blobsha1_imagesha"] = empty_column

    changeimage_base.to_csv(
        "./input/images_forceupdate_blob_baseload.adb.csv", index=False)
    # with a modified id, the id will change, but not the images.
    changeimage_modified.to_csv(
        "./input/images_forceupdate_updates.adb.csv", index=False)


def generate_sparse_add(multiplier):
    # generate base load, small images.
    image_count = 100
    img_gen = ImageGenerator(count=image_count, output="input/images/spare_add_image_%%.png", image_type="png",
                             size=(32, 32), manifest="input/sparse_add_image_list.csv")
    img_gen.run()
    img_df = pd.read_csv("input/sparse_add_image_list.csv", header=None)
    licence_count = 2
    multiplier = multiplier // 2
    img_ids = [i for i in range(image_count)] * multiplier
    license = [x for x in range(licence_count)] * multiplier

    def filemap(file_num):
        return remove_prefix(img_df.iat[file_num % image_count, 0], "input/")
    images  = list(product(img_ids, license))
    prop_change_count = len(images) // 10
    id_range = 500000000
    df = pd.DataFrame(images, columns=['img_id', 'license'])
    ids      = random.sample(range(id_range), len(images))
    age      = [int(100 * random.random()) for i in range(len(images))]
    height   = [float(200 * random.random()) for i in range(len(images))]
    dog      = [x > 100 for x in height]
    date_cap = [datetime.now().isoformat() for x in range(len(images))]
    version_id = [1 for x in range(len(images))]

    df.insert(0, "filename", df['img_id'].apply(filemap))
    df["id"]       = ids
    df["age"]      = age
    df["height"]   = height
    df["has_dog"]  = dog
    df["date:date_captured"] = date_cap
    df["constraint_id"] = ids
    df["version_id"] = version_id
    df["updateif_<version_id"] = version_id
    # create load with first half
    half = len(df) / 2
    # won't work with uneven, since we are going to expect to be able to double number to check.
    assert(int(half) == half)
    df.head(int(half)).to_csv(
        "./input/images_sparseload_base.adb.csv", index=False)

    # then load with all, which half will not be sent, as they exist.
    df.to_csv("./input/images_sparseload_full.adb.csv", index=False)


def main(params):

    persons = generate_person_csv(params.multiplier)
    blobs   = generate_blobs_csv()
    images  = generate_images_csv(int(params.multiplier / 2))
    s3_imgs = generate_http_images_csv("input/sample_http_urls.csv")
    s3_imgs = generate_s3_images_csv("input/sample_s3_urls.csv")
    generate_gs_images_csv("input/sample_gs_urls")
    connect = generate_connections_csv(persons, images)
    bboxes  = generate_bboxes_csv(images)
    bboxes_constraints = generate_bboxes_constraints_csv(images)

    desc_name = ["setA", "setB", "setC", "setD", "setE", "setF"]
    desc_dims = [2048, 1025, 2048, 1025, 2048, 1025]    # yes, 1025
    generate_descriptorset(desc_name, desc_dims)
    for name, dims in zip(desc_name, desc_dims):
        generate_descriptors(images, name, dims)

    generate_update_person()
    generate_partial_load()
    generate_update_image(params.multiplier)
    generate_update_image_fail(params.multiplier)
    generate_sparse_add(params.multiplier)
    generate_forceimage_load(params.multiplier)


def get_args():
    obj = argparse.ArgumentParser()

    # Run Config
    obj.add_argument('-multiplier', type=int, default=10)

    params = obj.parse_args()

    return params


if __name__ == "__main__":
    args = get_args()
    main(args)



# mark_checker.py - checks if a mark is enabled
from pytest import Session, Config, Item

local_config = None


def store_config(config):
    global local_config
    local_config = config


def is_enabled(test_mark):
    global local_config
    # create a config just using commandline args.
    c = local_config.fromdictargs(dict(), local_config.invocation_params.args)
    #raise Exception( str(c.getini("filterwarnings")) + " vs " + str(local_config.getini("filterwarnings")))
    s = Session.from_config(c)
    test_item = Item.from_parent(s, name= f"modulehere_test_{test_mark}")
    required = 0
    # add marker with given mark to this test.
    if isinstance(test_mark, (list, tuple)):
        for iter_mark in test_mark:
            test_item.add_marker(iter_mark)
        required = len(test_mark)
    else:
        test_item.add_marker(test_mark)
        required = 1
    items = [test_item]
    # pytest/mark modifies items based on supplied args.
    local_config.hook.pytest_collection_modifyitems(
        session=s, config=c, items=items)
    # if item is remaining, test would be run.
    return len(items) == 1



from __future__ import annotations
from PIL import Image, ImageDraw, ImageFont
import subprocess
import argparse
import re
import random
import hashlib
from colorsys import rgb_to_hsv, hsv_to_rgb


class OutputFilePath():
    def __init__(self, path):
        self.path = path

    def __str__(self):
        return f"{self.path}"

    def get_file(self, ext, number, zerofill=0):
        has_subst = False

        split = self.path.split("%%")
        stot = len(split)
        # ensure substituion and number are included.
        if stot > 1:
            final = []
            for i in range(stot - 1):
                final = final + [split[i], str(number).zfill(zerofill)]
            file = "".join(final + [split[stot - 1]])
            if not self.path.endswith(ext):
                file = file + ext

        else:
            file = self.path
            if self.path.endswith(ext):
                file = self.path[:-len(ext)]
            file = {file} + str(number).zfill(zerofill) + ext

        return file

    @classmethod
    def parse(cls, input_str: str) -> OutputFilePath:
        return OutputFilePath(input_str)


class ImageSize():
    def __init__(self, width, height):
        self.width = width
        self.height = height

    def size(self):
        return (self.width, self.height)

    def __str__(self):
        return f"{self.width}x{self.height}"

    @classmethod
    def parse(cls, input_str: str) -> ImageSize:
        width = 0
        height = 0
        fmt = re.compile("(^\d+)\s*(x|X)\s*(\d+)$")

        try:
            result = fmt.match(input_str)
            width = int(result.group(1))
            height = int(result.group(3))
        except:
            print("no")
            raise argparse.ArgumentTypeError(
                f"{input_str} is not a valid image size; requires width x height")

        return ImageSize(width, height)


fnt = None


def load_preferred_font(name, size):
    global fnt
    if fnt is not None:
        return fnt
    try:
        fnt = ImageFont.truetype(name, size)
    except OSError:
        best_font = subprocess.run(
            ["fc-match", "Sans"], stdout=subprocess.PIPE, text=True)
        if best_font.returncode != 0:
            raise Exception("Can't find a font")
        fc_output = best_font.stdout
        font = fc_output.split(":")[0]
        fnt = ImageFont.truetype(font, size)
    return fnt


def complementary_color(orig_color):
    hsv = rgb_to_hsv(*orig_color)
    # rotate hue and invert value
    return tuple(int(v) for v in hsv_to_rgb((hsv[0] + 0.5) % 1, hsv[1], abs(1 - hsv[2])))


def getoptions():
    parser = argparse.ArgumentParser()
    parser.add_argument('-c', '--count', type=int, default=1)
    parser.add_argument('-s', '--size', type=ImageSize.parse,
                        default=ImageSize(256, 256))
    parser.add_argument('-o', '--output', type=OutputFilePath.parse,
                        default=OutputFilePath("/tmp/generated_image%%"))
    parser.add_argument('-z', '--zerofill', type=int, default=0)
    parser.add_argument('-t', '--imagetype',
                        choices=['jpg', 'png'], default='png')
    parser.add_argument('-m', '--manifest', type=str, default=None)
    parser.add_argument('-a', '--append_text', type=str, default="")
    return parser.parse_args()


if __name__ == '__main__':
    params = getoptions()
    prog = ImageGenerator()
    prog.run(params)


class ImageGenerator:
    def __init__(self, count=1, size=(256, 256), output="/tmp/generated_image%%", zerofill=0, image_type="png", manifest=None, append_text=""):
        self.write_log = []
        try:
            output_path = OutputFilePath(output)
        except Exception as e:
            raise Exception(f"Bad output path {str(output)}: {e} ")
        try:
            image_size = ImageSize(size[0], size[1])
        except Exception as e:
            raise Exception(f"Bad image size {str(size)}: {e} ")

        options_map = {
            "count": count,
            "size": image_size,
            "output": output_path,
            "zerofill": zerofill,
            "imagetype": image_type,
            "manifest": manifest,
            "append_text": append_text
        }

        class Options:
            def __init__(self, opt_in):
                for key in opt_in:
                    setattr(self, key, opt_in[key])
        self.options = Options(options_map)

    def draw_and_save(self, of, ot, os, text):
        # choose color based on hash output
        bg_color = tuple(random.randint(0, 255) for _ in range(3))
        fg_color = complementary_color(bg_color)
        color_white = (255, 255, 255, 255)
        color_red = (255, 0, 0, 255)
        canvas = Image.new("RGBA", os.size(), bg_color)
        fnt = load_preferred_font('DejaVuSans-Bold.ttf', 16)
        d = ImageDraw.Draw(canvas)
        d.text((10, 10), text, font=fnt, fill=fg_color)
        canvas.save(of, ot)
        self.write_log.append(of)
        print(f"* Saved {of} of size {os} of type {ot} with text {text}")

    def run(self, input_params=None):
        params = self.options if input_params is None else input_params

        images = "image" if params.count == 1 else "images"
        print(f"Creating {params.count} {images}")
        ext = "." + params.imagetype
        # set random based on filename, but stable between runs.
        random.seed(
            int(hashlib.md5(params.output.get_file(ext, 0, params.zerofill).encode()).hexdigest(), 16))
        for i in range(0, params.count):
            self.draw_and_save(params.output.get_file(ext, i, params.zerofill),
                               params.imagetype, params.size, f"{i}{params.append_text}")

        if params.manifest:
            print(f"Writing manifest to {params.manifest}")
            with open(params.manifest, "w") as fp:
                for line in self.write_log:
                    fp.write(line + "\n")



import time
import logging
import ssl

from aperturedb.Connector import Connector

logger = logging.getLogger(__name__)


class TestSession():

    '''
        These check operation of the Connector Session
    '''

    def test_sessionRenew(self, db: Connector):
        '''
            Verifies that Session renew works
        '''

        try:
            # force session token expiry
            db.shared_data.session.session_token_ttl = 1
            logging.debug("Connected? {0}".format(
                "yes" if db.connected else "no"))

            logging.debug(
                "Session valid? {0}".format(
                    "yes" if db.shared_data.session.valid() else "no"))
            logging.debug("Valid length: {0}".format(
                db.shared_data.session.session_token_ttl))
            time.sleep(2)
            query = [{
                "FindImage": {
                    "results": {
                        "limit": 5
                    }
                }
            }]
            responses, blobs = db.query(query)
            logging.debug(responses)
            assert db.shared_data.session.valid() == True
        except Exception as e:
            print(e)
            print("Failed to renew Session")
            assert False

    def test_SSL_error_on_query(self, db: Connector, monkeypatch):
        # force session token expiry
        original_send_msg = db._send_msg
        count = 0

        # This will raise an SSL error on the first call to _send_msg (which is an refresh token)
        # The later calls will succeed.
        # When we try to renew the session, that attempts to take a lock on the session.
        # And this is the hang.
        def mock_send_msg(msg):
            nonlocal count
            count += 1
            logger.debug(count)
            if count == 1:
                raise ssl.SSLError("SSL Error")

            logger.debug(f"Cut {msg}")
            result  = original_send_msg(msg)
            logger.debug(result)
            return result

        monkeypatch.setattr(db, "_send_msg", lambda x: mock_send_msg(x))
        db.shared_data.session.session_token_ttl = 1
        time.sleep(2)
        query = [{
            "FindImage": {
                "results": {
                    "limit": 5
                }
            }
        }]
        responses, blobs = db.query(query)
        logging.debug(responses)
        assert db.shared_data.session.valid() == True



from __future__ import annotations
from typing import Callable
from aperturedb.ParallelQuery import ParallelQuery
import itertools
import logging
import numpy as np


logger = logging.getLogger(__name__)

# Turn on to debug constraints ( produces a lot of output )
DEBUG_CONSTRAINTS = True

# removes blobs from a list or tuple for pretty printing


def remove_blobs(item):
    if isinstance(item, list):
        item = list(map(remove_blobs, item))
    elif isinstance(item, tuple):
        item = list(map(remove_blobs, item))
    elif isinstance(item, bytes):
        item = "*BLOB*"
    return item


def gen_execute_batch_sets(base_executor, per_batch_response_handler: Callable = None):

    #
    # execute_batch_sets - executes multiple sets of queries with optional constraints on follow on sets
    #  query_set: list of queries that get passed to ParallelLoader's execute_batch
    #  blob_set:  list of blobs ( or list of list of blobs ) that get passed to execute_batch
    #  db: Connection object
    #  success_statutes: list of return values from ApertureDB to be considered 'success'
    #  response_handler: optional function to return after an execution ( not active yet )
    #  commands_per_query : list of how many commands each query has.
    #  blobs_per_query: list of how many blobs each query has.
    #  strict_response_validation: same as execute_batch.
    #
    # if blob_set is None, or an empty list, it is ignored.
    # if blob_set is a list, it will be given to the seed query
    # if blob_set is a list of lists, each inner list will be given to the inner
    #  execution
    #
    def execute_batch_sets(query_set, blob_set, db, success_statuses: list[int] = [0],
                           response_handler: Callable = None, commands_per_query: list[int] = -1,
                           blobs_per_query: list[int] = -1, strict_response_validation: bool = False):

        logger.info("Execute Batch Sets = Batch Size {0}  Comands Per Query {1} Blobs Per Query {2}".format(
            len(query_set), commands_per_query, blobs_per_query))

        batch_size = len(query_set)

        # test query set
        first_element = query_set[0]
        if not isinstance(first_element, list):
            logger.error("First Element not a list: {first_element}")
            raise Exception("Query set must be a list of lists")
        set_total = len(first_element)

        # Check if blobs are a simple array or nested array of blobs
        per_set_blobs = isinstance(blob_set, list) and len(
            blob_set) > 0 and isinstance(blob_set[0], list)

        # verify layout if a complex set
        if per_set_blobs:
            first_element_blobs = blob_set[0]
            first_query_blobs = first_element_blobs[0]
            # If someone is looking for info logging from PQS, it is likely that blobs are not being set properly.
            #  The wrapping of blobs in general can be confusing. Best suggestion is looking at a loader.
            logger.info("Blobs for first set = " +
                        str(remove_blobs(blob_set[0])))
            logger.info("First Blob for first set = " +
                        str(remove_blobs(blob_set[0][0])))
            if not isinstance(first_query_blobs, list):
                logger.error(
                    "Expected a list of lists for the first element's blob sets")
                expected_list = remove_blobs(first_query_blobs)
                logger.error(
                    f"Blob input for set is : {expected_list}, but should be a list")
                raise Exception(
                    "Could not determine blob strategy; first batch element doesn't have a list of blobs for the first query set. Are you missing a list wrapping in the CVS parser?")

            for set_i in range(len(first_element_blobs)):
                expected_set = first_element_blobs[set_i]
                if not isinstance(expected_set, list):
                    without_blobs = remove_blobs(expected_set)
                    logger.error(
                        f"Blob input for set {set_i} for first item is : {without_blobs}, but should be a list")
                    raise Exception(
                        f"Could not determine blob strategy; first batch element doesn't have a list of blobs for set {set_i}")
                if len(expected_set) != blobs_per_query[set_i]:
                    without_blobs = remove_blobs(expected_set)
                    logger.error(
                        f"Blob input for set {set_i} for first item is : {without_blobs}, but expecting a length of {blobs_per_query[set_i]}")
                    raise Exception(
                        f"Could not executed blob strategy; first batch element's list length doesn't match blob_per_query for the set {set_i} ( {len(expected_set)} != {blobs_per_query[set_i]} )")

        # define method for extracting the blobs
        def blob_filter(all_blobs, strike_list, set_nm): return all_blobs

        # set blob filter based on how blobs are passed to function
        if per_set_blobs:
            # if each query_set has blobs, we must extract out the n-th element in each blob set

            def set_blob_filter(all_blobs, strike_list, set_nm):
                # the list comprehension pulls out the blob set for the requested set
                # the blob set is then flattened as the query expects a flat array using blobs_per_query as the iterator
                # the flat list is them zipped with the strike list, which determines which blobs are unused
                # the filter checks if the blob is to be struc
                # the map pulls the remaining blobs out

                return list(map(lambda pair: pair[0],
                            filter(lambda pair: pair[1] is not True,
                                   itertools.zip_longest(
                                itertools.chain(*[blob_set[set_nm]
                                                for blob_set in all_blobs]),
                                strike_list)
                )
                ))
            blob_filter = set_blob_filter
        else:
            def first_only_blobs(all_blobs, strike_list, set_nm):
                if set_nm == 0:
                    # same as above, without the list comprehension and flattening.
                    return list(map(lambda pair: pair[0], filter(lambda pair: pair[1] is not False,   itertools.zip_longest(all_blobs, strike_list))))
                else:
                    return []
            blob_filter = first_only_blobs

        # start execution
        stored_results = {}
        db_results = []
        db_blobs = None
        for i in range(0, set_total):

            # total number of blobs in this set
            # passing in [] allows for no strikes, so maximum size.
            blobs_this_set = len(blob_filter(blob_set, [], i))
            expected_blobs = blobs_per_query[i] * batch_size
            logger.info(
                f"Set {i}: Commands per query = {commands_per_query[i]}, Blobs per query = {blobs_per_query[i]}"
            )
            if blobs_this_set != expected_blobs:
                logger.error(
                    f"Set {i}: Expected {expected_blobs} blobs, but filter is returning {blobs_this_set}"
                )

            # now we determine if the executing set has a constraint
            # in the head

            # allowed layouts for commands other than the seed command
            # { "cmd" : {} } -> standard single command
            # [{ "cmd1": {}, "cmd2} : {}] -> standard multiple command
            # [{ "constraint" : {} , { "cmd" : {} }] -> constraint with a single command
            # [{ "constraints: {} , [{"cmd1" : {} }, {"cmd2": {} }]] -> constraint with multiple command

            known_constraint_keys = ["results", "apply"]
            constraints = None
            # define filter to optionally remove constraints when passing to executor
            def query_filter(entity, entity_results): return entity[i]

            if i != 0 and isinstance(first_element[i], list) and len(first_element[i]) > 0:
                set_component = first_element[i][0]
                if not isinstance(set_component, list) and any([k in known_constraint_keys for k in set_component.keys()]):
                    constraints = set_component

            if constraints is not None:
                # operations for constraints
                def check_apply_constraint(test_op, item_a, item_b):
                    if DEBUG_CONSTRAINTS:
                        logger.debug(
                            f"check_apply_constraint {test_op} {item_a} {item_b}")
                    if test_op == "==":
                        return item_a == item_b
                    elif test_op == ">":
                        return item_a > item_b
                    elif test_op == "<":
                        return item_a < item_b
                    elif test_op == "!=":
                        return item_a != item_b
                    else:
                        raise Exception(
                            f"Unhandled constraint {test_op} in check_apply_constraint")

                known_constraint_keywords = ['results', 'apply']
                # function called for each row in the set to decide if a query is executed

                def constraint_filter(single_line, single_results):
                    current_constraints = single_line[i][0]

                    if DEBUG_CONSTRAINTS:
                        logger.debug(f"constraint = {current_constraints}")
                        logger.debug(f"results = {single_results}")

                    if 'results' in current_constraints:
                        result_constraints = current_constraints['results']
                        passed_all_constraints = True
                        for result_number in result_constraints:

                            if len(single_results) < result_number or single_results[result_number] is None:
                                # in theory here we have two possibilities: a user can have a correctly formed constraint which didn't execute by design
                                # ( which is what process here )
                                #  or they can have a mis-configured constraint, which is harder to detect.
                                if DEBUG_CONSTRAINTS:
                                    logger.debug(
                                        f"Failed applying constraints, requested results from operation {result_number}, but none existed ( Query didn't run )")
                                passed_all_constraints = False
                                break

                            # query this constraint references
                            prev_query = single_line[result_number]

                            if DEBUG_CONSTRAINTS:
                                logger.debug(
                                    f"prev_query = {single_line[result_number]}")

                            if isinstance(prev_query, dict):
                                prev_query_cmd = [
                                    cmd for cmd in prev_query.keys()][0]
                            elif isinstance(prev_query, list) and isinstance(prev_query[0], dict) \
                                    and all(map(lambda k: k in known_constraint_keywords, prev_query[0].keys()))  \
                                    and isinstance(prev_query[1], dict):
                                prev_query_cmd = [
                                    cmd for cmd in prev_query[1].keys()][0]
                            else:
                                raise Exception(
                                    f"Contraints only implemented with with single queries; query {result_number} not single item.")

                            target_results = single_results[result_number][prev_query_cmd]
                            target_constraints = result_constraints[result_number]
                            for result_item in target_constraints:
                                test = target_constraints[result_item]
                                if not result_item in target_results:
                                    logger.debug(
                                        f"failed results = {target_results}")
                                    raise Exception(
                                        f"Failed applying constraints, requested result '{result_item}' from operation {result_number}, but none exited")
                                # if constraint passes, apply:
                                if not check_apply_constraint(test[0], target_results[result_item], test[1]):
                                    passed_all_constraints = False
                                    break

                        if passed_all_constraints:
                            return single_line[i][1]
                        else:
                            return None
                    elif 'apply' in current_constraints:
                        # apply means run the line
                        return single_line[i][1]
                    else:
                        raise Exception(
                            f"incorrectly formatted constraint; no known conatraint action")

                query_filter = constraint_filter

            local_success_statuses = [0, 2]

            # queries are by row first, so we run query_filter on each query
            # we pass the entire row's data, then we retrieve all of the stored results for that row

            queries = [query_filter(query_set[row_num], [stored_results[res_grp][row_num] for res_grp in stored_results])
                       for row_num in range(0, len(query_set))]

            # finally, we remove queries that were reduced to None so the base executor doesn't have to deal with them
            # and unwrap multiple commands per element into a flat list; map allows one command to process either type
            #  by wrapping all in a list.
            executable_queries = list(itertools.chain(*
                                                      map(lambda q: q if isinstance(q, list) else [q],
                                                          filter(lambda q: q is not None, queries))
                                                      ))

            # produce strike list based on queries that are none
            blob_strike_list = list(map(lambda q: q is None, queries))

            # filter out struck blobs
            used_blobs = filter(lambda b: b is not None,
                                blob_filter(blob_set, blob_strike_list, i))

            # TODO: add wrapped response_handler.
            if response_handler != None:
                logger.warning(
                    "ParallelQuerySet does not yet support a response_handler which will identify which set is being worked on")
            if len(executable_queries) > 0:
                result_code, db_results, db_blobs = base_executor(executable_queries, used_blobs,
                                                                  db, local_success_statuses,
                                                                  None, commands_per_query[i], blobs_per_query[i], strict_response_validation=strict_response_validation)
            else:
                logger.info(
                    f"Skipped executing set {i}, no executable queries")

            if result_code == 1:
                logger.error(
                    f"Ran into error on set {i} in ParallQuerySet, unable to continue")
                return 1, db_results, db_blobs

            # expand results so queries that didn't run show up as a None
            results_off = 0

            def insert_empty_results(result_value):
                nonlocal results_off
                if result_value is None:
                    return None
                else:
                    results_off += 1
                    return db_results[(results_off - 1)]

            stored_results[i] = [insert_empty_results(q) for q in queries]

        return 0, db_results, db_blobs  # end execute_batch_sets
    return execute_batch_sets


class ParallelQuerySet(ParallelQuery):
    """
    **Parallel and Batch Set Multi-Querier for ApertureDB**
    This class provides the mechanism to run multiple queries over a single csv.
     Per-query actions are done by ParallelQuery.
    """

    def __init__(self, db, dry_run=False):

        super().__init__(db, dry_run)

        self.base_batch_command = self.batch_command

    def verify_generator(self, generator):
        # first level should be grouping of commands
        # first cmd should have a list of query sets
        if isinstance(generator[0], list) or isinstance(generator[0], tuple):
            cmd = generator[0]
            if isinstance(generator[0][0], list):
                return True

        logger.error(
            f"Could not determine query structure from:\n{generator[0]}")
        logger.error(type(generator[0]))
        return False

    def do_batch(self, db, data):
        """
        This is an override of ParallelQuery.do_batch.

        This is the per-worker function which is the entry-point to a unit of work.
        """
        if not hasattr(self.generator, "commands_per_query"):
            raise Exception(
                "Generator must specify commands per query for ParallelQuerySet")
        if not hasattr(self.generator, "blobs_per_query"):
            raise Exception(
                "Generator must specify blobs per query for ParallelQuerySet")
        self.commands_per_query = self.generator.commands_per_query
        self.blobs_per_query = self.generator.blobs_per_query
        set_response_handler = None
        if hasattr(self.generator, "set_response_handler") and callable(self.generator.set_response_handler):
            set_response_handler = self.generator.set_response_handler
        self.batch_command = gen_execute_batch_sets(
            self.base_batch_command, set_response_handler)

        ParallelQuery.do_batch(self, db, data)

    def print_stats(self):

        times = np.array(self.times_arr)
        total_queries_exec = len(times)

        # Todo - implement per-set data.

        print("============ ApertureDB Parallel Query Set Stats ============")
        print(f"Total time (s): {self.total_actions_time}")
        print(f"Total queries executed: {total_queries_exec}")

        if total_queries_exec == 0:
            print("All queries failed!")

        else:
            mean = np.mean(times)
            std  = np.std(times)
            tp = 1 / mean * self.numthreads

            print(f"Avg Query time (s): {mean}")
            print(f"Query time std: {std}")
            print(f"Avg Query Throughput (q/s): {tp}")

            i_tp = self.total_actions / self.total_actions_time
            print(
                f"Overall insertion throughput ({self.type}/s): {i_tp if self.error_counter == 0 else 'NaN'}")

            if self.error_counter > 0:
                err_perc = 100 * self.error_counter / total_queries_exec
                print(f"Total errors encountered: {self.error_counter}")
                print(f"Errors (%): {err_perc}")

        print("=========================================================")



import pytest
import mark_checker
from aperturedb.Connector import Connector
from aperturedb.ConnectorRest import ConnectorRest
from aperturedb.ParallelLoader import ParallelLoader
from aperturedb.ParallelQuerySet import ParallelQuerySet
from aperturedb.BlobDataCSV import BlobDataCSV
from aperturedb.EntityDataCSV import EntityDataCSV
from aperturedb.EntityUpdateDataCSV import SingleEntityUpdateDataCSV
from aperturedb.ConnectionDataCSV import ConnectionDataCSV
from aperturedb.DescriptorSetDataCSV import DescriptorSetDataCSV
from aperturedb.DescriptorDataCSV import DescriptorDataCSV
from aperturedb.ImageDataCSV import ImageDataCSV, ImageUpdateDataCSV, ImageForceNewestDataCSV, ImageSparseAddDataCSV
from aperturedb.BBoxDataCSV import BBoxDataCSV
from aperturedb.Constraints import Constraints
from aperturedb.Entities import Entities
from aperturedb.Query import Query
from aperturedb.Utils import Utils

import dbinfo

# These are test fixtures and can be used in any
# pytest tests as function parmeters with same names.


def pytest_generate_tests(metafunc):
    if "db" in metafunc.fixturenames:
        metafunc.parametrize("db", [
            {"db": Connector(
                host = dbinfo.DB_TCP_HOST,
                port = dbinfo.DB_TCP_PORT,
                user = dbinfo.DB_USER,
                password = dbinfo.DB_PASSWORD,
                use_ssl = True)},
            {"db": ConnectorRest(
                host = dbinfo.DB_REST_HOST,
                port = dbinfo.DB_REST_PORT,
                user = dbinfo.DB_USER,
                password = dbinfo.DB_PASSWORD,
                use_ssl = False
            )}
        ], indirect=True, ids=["TCP", "HTTP"])
    if all(func in metafunc.fixturenames for func in ["insert_data_from_csv", "modify_data_from_csv"]) and \
            metafunc.module.__name__ in ["test.test_Data"]:
        metafunc.parametrize("insert_data_from_csv,modify_data_from_csv", [
                             [True, True], [False, False]], indirect=True, ids=["with_dask", "without_dask"])
    elif "insert_data_from_csv" in metafunc.fixturenames and metafunc.module.__name__ in \
            ["test.test_Data"]:
        metafunc.parametrize("insert_data_from_csv", [
                             True, False], indirect=True, ids=["with_dask", "without_dask"])
    elif "modify_data_from_csv" in metafunc.fixturenames and metafunc.module.__name__ in \
            ["test.test_Data"]:
        metafunc.parametrize("modify_data_from_csv", [
                             True, False], indirect=True, ids=["with_dask", "without_dask"])


@pytest.fixture()
def db(request):
    db = request.param['db']
    utils = Utils(db)
    assert utils.remove_all_objects()
    return db


def check_response_regressions(queries, input_blobs, responses, output_blobs):
    # Check that responses have no blobs
    first_command = list(responses[0].keys())[0]
    assert "blobs_start" not in responses[0][
        first_command], f"responses[0]={responses[0]}"


@pytest.fixture()
def insert_data_from_csv(db, request):
    """
    A helper function that processes various .csv files supported
    by aperturedb, and maps a corresponding DataCSV class that can be
    used to parse semantics of the .csv file
    """
    def insert_data_from_csv(in_csv_file, rec_count=-1, expected_error_count=0, loader_result_lambda=None):
        if rec_count > 0 and rec_count < 80:
            request.param = False
            print("Not enough records to test parallel loader. Using serial loader.")
        file_data_pair = {
            "./input/persons.adb.csv": EntityDataCSV,
            "./input/images.adb.csv": ImageDataCSV,
            "./input/connections-persons-images.adb.csv": ConnectionDataCSV,
            "./input/bboxes.adb.csv": BBoxDataCSV,
            "./input/blobs.adb.csv": BlobDataCSV,
            "./input/descriptorset.adb.csv": DescriptorSetDataCSV,
            "./input/setA.adb.csv": DescriptorDataCSV,
            "./input/setB.adb.csv": DescriptorDataCSV,
            "./input/s3_images.adb.csv": ImageDataCSV,
            "./input/http_images.adb.csv": ImageDataCSV,
            "./input/bboxes-constraints.adb.csv": BBoxDataCSV,
            "./input/gs_images.adb.csv": ImageDataCSV,
            "./input/images_to_modify.adb.csv": ImageDataCSV,
            "./input/images_to_modify.adb.csv": ImageDataCSV,
            "./input/persons-exist-base.adb.csv": EntityDataCSV,
            "./input/persons-some-exist.adb.csv": EntityDataCSV
        }
        use_dask = False
        if hasattr(request, "param"):
            use_dask = request.param
        data = file_data_pair[in_csv_file](in_csv_file, use_dask=use_dask,
                                           blobs_relative_to_csv=True)

        setattr(data, "response_handler", check_response_regressions)
        data.strict_response_validation = True

        if rec_count != -1:
            data = data[:rec_count]

        loader = ParallelLoader(db)
        loader.ingest(data, batchsize=99,
                      numthreads=8,
                      stats=True,
                      )

        # make sure all indices are present
        if hasattr(data, "get_indices"):
            expected_indices = data.get_indices()
            observed_indices = loader.get_existing_indices()
            for tp, classes in expected_indices.items():
                for cls, props in classes.items():
                    for prop in props:
                        assert prop in observed_indices[tp][cls]

        assert loader.error_counter == 0
        assert len(data) - \
            loader.get_suceeded_queries() == expected_error_count
        if loader_result_lambda is not None:
            loader_result_lambda(loader, data)
        assert len(data) - \
            loader.get_suceeded_queries() == expected_error_count
        if loader_result_lambda is not None:
            loader_result_lambda(loader, data)

        # Preserve loader so that dask manager is not auto deleted.
        # ---------------
        # Previously, dask's cluster and client were entirely managed in a context
        # insede dask manager. A change upstream broke that, and now we keep the DM
        # in scope just so that we can do computations after ingestion, as those
        # things are lazily evaluated.
        return data, loader

    return insert_data_from_csv


class UpdatePersonEntityDataCSV(SingleEntityUpdateDataCSV):
    def __init__(self, filename, **kwargs):
        super().__init__("Person", filename, **kwargs)


@pytest.fixture()
def modify_data_from_csv(db, request):
    def modify_data_from_csv(in_csv_file, rec_count=-1):
        if rec_count > 0 and rec_count < 80:
            request.param = False
            print("Not enough records to test parallel loader. Using serial loader.")
        file_data_pair = {
            "./input/persons-update.adb.csv": UpdatePersonEntityDataCSV,
            "./input/persons-update-oldversion.adb.csv": UpdatePersonEntityDataCSV,
            "./input/persons-update-newversion.adb.csv": UpdatePersonEntityDataCSV,
            "./input/persons-update-olderage.adb.csv": UpdatePersonEntityDataCSV,
            "./input/images_updateif_baseload.adb.csv": ImageUpdateDataCSV,
            "./input/images_updateif_mixednew.adb.csv": ImageUpdateDataCSV,
            "./input/images_updateif_fail_baseload.adb.csv": ImageUpdateDataCSV,
            "./input/images_updateif_fail_updates.adb.csv": ImageUpdateDataCSV,
            "./input/images_forceupdate_baseload.adb.csv": ImageForceNewestDataCSV,
            "./input/images_forceupdate_mixednew.adb.csv": ImageForceNewestDataCSV,
            "./input/images_forceupdate_fail_base.adb.csv": ImageForceNewestDataCSV,
            "./input/images_forceupdate_fail_updates.adb.csv": ImageForceNewestDataCSV,
            "./input/images_forceupdate_blob_baseload.adb.csv": ImageForceNewestDataCSV,
            "./input/images_forceupdate_updates.adb.csv": ImageForceNewestDataCSV,
            "./input/images_sparseload_base.adb.csv": ImageSparseAddDataCSV,
            "./input/images_sparseload_full.adb.csv": ImageSparseAddDataCSV
        }
        use_dask = False
        if hasattr(request, "param"):
            use_dask = request.param
        data = file_data_pair[in_csv_file](
            in_csv_file, use_dask=use_dask, blobs_relative_to_csv=True)
        if rec_count != -1:
            data = data[:rec_count]

        loader = ParallelQuerySet(db)
        loader.query(data, batchsize=99,
                     numthreads=1,
                     stats=True,
                     )
        assert loader.error_counter == 0
        return data, loader

    return modify_data_from_csv


@pytest.fixture()
def utils(db):
    return Utils(db)


@pytest.fixture()
def images(insert_data_from_csv):
    return insert_data_from_csv("./input/images.adb.csv")


@pytest.fixture()
def retired_persons(db, insert_data_from_csv, utils):
    utils.remove_entities(class_name="Person")
    loaded = insert_data_from_csv("./input/persons.adb.csv")
    constraints = Constraints()
    constraints.greaterequal("age", 60)
    retired_persons = Entities.retrieve(
        db, spec=Query.spec(
            with_class="Person", constraints=constraints))
    return retired_persons


@pytest.fixture()
def config():
    config = {
        "first": {
            "host": "localhost",
            "port": 55555,
            "username": "admin",
            "password": "admin",
            "name": "first",
            "use_ssl": True,
            "use_rest": False
        },
        "active": "first"
    }
    return config


def pytest_configure(config):
    mark_checker.store_config(config)



import os
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

OPENCV_VERSION = os.getenv("OPENCV_VERSION")

install_requires = ['scikit-image', 'image', 'requests', 'boto3',
                    'numpy', 'matplotlib', 'pandas', 'kaggle', 'google-cloud-storage',
                    'dask[complete]', 'ipywidgets', 'pydantic', 'devtools', 'typer[all]',
                    # Pin to the bridge version.
                    # https://github.com/tensorflow/tensorflow/issues/60320
                    'protobuf==3.20.3',
                    # Pinning this to be able to install google-cloud-bigquery
                    'grpcio-status==1.48.2',
                    # Pinning this to resolve test errors temporarily
                    'ipywidgets==8.0.4',
                    'keepalive-socket==0.0.1'
                    ]
if OPENCV_VERSION is None:
    install_requires.append('opencv-python')

setuptools.setup(
    name="aperturedb",
    version="0.4.10",
    description="ApertureDB Client Module",
    install_requires=install_requires,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/aperture-data/aperturedb-python",
    license="Apache",
    packages=setuptools.find_packages(),
    python_requires='>=2.6, !=3.0.*, !=3.1.*, !=3.2.*, <4',
    author="Luis Remis",
    author_email="luis@aperturedata.io",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    entry_points = {
        'console_scripts': ['adb=aperturedb.cli.adb:app'],
    }
)
